#ifndef AABB_POLYHEDRAL_ORACLE_H
#define AABB_POLYHEDRAL_ORACLE_H

#include <boost/static_warning.hpp>
#include <utility>
#include <CGAL/iterator.h>

#include "AABB_tree.h"


// Modified version of CGAL::AABB_polyhedral_oracle to be independent of a polyhedron class.
template <class TriangleKernel,           ///< Triangles' kernel
          class TriangleIterator,         ///< Value type must be Triangle_3
          class AABBTree_kernel = CGAL::Simple_cartesian<typename TriangleKernel::FT>
                                          ///< Local kernel for fast bounding box intersections
>                                          
//template <class Polyhedron, class TriangleKernel, class AABBTree_kernel>
class AABB_polyhedral_oracle 
//: public Polyhedron
{
    typedef AABB_polyhedral_oracle Self;
    
public:
    //typedef typename TriangleKernel::FT FT;
    //typedef typename TriangleKernel::Ray_3 Ray_3;
    //typedef typename TriangleKernel::Line_3 Line_3;
    typedef typename TriangleKernel::Point_3 Point_3;
    //typedef typename TriangleKernel::Segment_3 Segment_3;

    typedef Self Surface_mesher_traits_3;
    typedef Point_3 Intersection_point;
    typedef Self Surface_3;

    // AABB tree
    typedef AABB_tree<TriangleKernel, TriangleIterator, AABBTree_kernel> Tree;
    typedef typename Tree::Point_with_triangle Point_with_triangle;

private:
    Tree *m_pTree;

public:
    Tree* tree() const{ return m_pTree; }

public:
    // constructor
    AABB_polyhedral_oracle()
    {
        m_pTree = NULL;
    }
    AABB_polyhedral_oracle(Tree *pTree)
    {
        m_pTree = pTree;
    }
    //AABB_polyhedral_oracle(const AABB_polyhedral_oracle& oracle)
    //{
    //    m_pTree = oracle.tree();
    //}

    //class Intersect_3;
    //friend class Intersect_3;

    class Intersect_3 {
        const Self& self;
    public:
        Intersect_3(const Self& self) : self(self)
        {
        }

        template<class SegmentType>    ///< tested with Segment_3 and Ray_3
        CGAL::Object operator()(const Surface_3& surface, const SegmentType& segment) const
        {
            Point_with_triangle pwh;
            if(surface.tree()->first_intersection(segment,pwh))
                return CGAL::make_object(pwh.first);
            else
                return CGAL::Object();
        }

        //Object operator()(const Surface_3& surface, const Ray_3& ray) const
        //{
        //    Converter convert;
        //    BConverter bconvert;
        //    Point_with_triangle pwh;
        //    if(surface.tree()->first_intersection(convert(ray),pwh))
        //        return make_object(bconvert(pwh.first));
        //    else
        //        return Object();
        //}

        //Object operator()(const Surface_3& surface, const Line_3& line) const
        //{
        //    Converter convert;
        //    BConverter bconvert;
        //    Point_with_triangle pwh;
        //    if(surface.tree()->first_intersection(convert(line),pwh))
        //        return make_object(bconvert(pwh.first));
        //    else
        //        return Object();
        //}
        
    }; // class Intersect_3

    Intersect_3 intersect_3_object() const
    {
        return Intersect_3(*this);
    }

    //class Construct_initial_points;
    //friend class Construct_initial_points;

    class Construct_initial_points
    {
        const Self& self;
    public:
        Construct_initial_points(const Self& self) : self(self)
        {
        }

        template <typename OutputIteratorPoints>
        OutputIteratorPoints operator() (const Surface_3& surface, 
                                         OutputIteratorPoints out, 
                                         int n) const 
        {
            // TODO (with visitor)
            // std::cout << "construct initial point set" << std::endl;
            // *out++= p;
            return out;
        }
        
    }; // class Construct_initial_points

    Construct_initial_points construct_initial_points_object() const
    {
        return Construct_initial_points(*this);
    }

    template <class P>
    bool is_in_volume(const Surface_3& surface, const P& p)
    {
        std::cout << "DEBUG: call is in volume" << std::endl;
        return true;
    }
    
}; // end class AABB_polyhedral_oracle


#endif // AABB_POLYHEDRAL_ORACLE_H
